// This file is part of the adequate Debian-native package, and is available
// under the Expat license. For the full terms please see debian/copyright.

package main

import (
	"fmt"
	"log"
	"regexp"
	"slices"
)

const (
	MISSING_ALTERNATIVE_TAG = "missing-alternative"
)

type altErr struct {
	pkg  string
	vpkg string
}

func (c altErr) Error() string {
	if c == (altErr{}) {
		return ""
	}
	return fmt.Sprintf("%s: %s %s", c.pkg, MISSING_ALTERNATIVE_TAG, c.vpkg)
}

type altChecker struct {
	emit bool
}

func newAltChecker(tags tagFilter) altChecker {
	return altChecker{
		emit: tags.shouldEmit(MISSING_ALTERNATIVE_TAG),
	}
}

func (cc altChecker) check(pkgProvides map[string]map[string]bool, pkg2files map[string][]string) []altErr {
	if !cc.emit {
		return nil
	}

	vpkgs := []string{"x-window-manager", "x-terminal-emulator"}
	providers := make(map[string][]string)
	for pkg, provides := range pkgProvides {
		for _, vpkg := range vpkgs {
			if provides[vpkg] {
				providers[vpkg] = append(providers[vpkg], pkg)
			}
		}
	}
	if len(providers) == 0 {
		return nil
	}

	var tags []altErr
	altMap := makeAltMap(vpkgs)
	for vpkg, pkgs := range providers {
		var registeredPaths []string
		for _, k := range altMap[vpkg] {
			registeredPaths = append(registeredPaths, k)
		}
		for _, pkg := range pkgs {
			var found bool
			for _, f := range pkg2files[pkg] {
				if slices.Contains(registeredPaths, f) {
					found = true
					break
				}
			}
			if !found {
				tags = append(tags, altErr{
					pkg:  pkg,
					vpkg: vpkg,
				})
			}
		}
	}

	return tags
}

func makeAltMap(interestingAlts []string) map[string][]string {
	lines, err := runCommand([]string{"update-alternatives", "--get-selections"})
	if err != nil {
		log.Fatal("update-alternatives --get-selections: ", err)
	}

	// sample input:
	// 	cpp                            auto     /usr/bin/cpp
	// TODO: handle missing 3rd column (#857265).
	re := regexp.MustCompile("^(" + nonSpaceToken + ")" + spaceToken + nonSpaceToken + spaceToken + nonSpaceToken + "$")
	seenAlts := make(map[string]bool)
	for _, line := range lines {
		m := re.FindStringSubmatch(line)
		if len(m) != 2 {
			log.Fatalf("Unexpected output from update-alternatives --get-selections: %q", line)
		}
		seenAlts[m[1]] = true
	}

	m := make(map[string][]string)
	for _, alt := range interestingAlts {
		if !seenAlts[alt] {
			continue
		}

		alts, err := runCommand([]string{"update-alternatives", "--list", alt})
		if err != nil {
			log.Fatal("update-alternatives --list: ", err)
		}
		m[alt] = append(m[alt], alts...)
	}

	return m
}
